# --------------------------------------------------------
# References:
# https://github.com/jxhe/unify-parameter-efficient-tuning
# --------------------------------------------------------

import math
import torch
import torch.nn as nn


class Adapter(nn.Module):
    def __init__(
        self,
        config=None,
        d_model=768,
        bottleneck=None,
        dropout=0.0,
        init_option="lora",
        adapter_scalar="1.0",
        adapter_layernorm_option="none"
    ):
        super().__init__()
        self.n_embd = config.d_model if d_model is None else d_model
        self.down_size = config.attn_bn if bottleneck is None else bottleneck

        #_before
        self.adapter_layernorm_option = adapter_layernorm_option

        self.adapter_layer_norm_before = None
        if adapter_layernorm_option == "in" or adapter_layernorm_option == "out":
            self.adapter_layer_norm_before = nn.LayerNorm(self.n_embd)

        if adapter_scalar == "learnable_scalar":
            self.scale = nn.Parameter(torch.ones(1))
        else:
            self.scale = float(adapter_scalar)

        self.down_proj = nn.Linear(self.n_embd, self.down_size)
        self.non_linear_func = nn.ReLU()
        self.up_proj = nn.Linear(self.down_size, self.n_embd)

        self.dropout = dropout
        if init_option == "bert":
            raise NotImplementedError
        elif init_option == "lora":
            with torch.no_grad():
                nn.init.kaiming_uniform_(self.down_proj.weight, a=math.sqrt(5))
                nn.init.zeros_(self.up_proj.weight)
                nn.init.zeros_(self.down_proj.bias)
                nn.init.zeros_(self.up_proj.bias)

    def forward(self, x, add_residual=True, residual=None):
        residual = x if residual is None else residual
        if self.adapter_layernorm_option == 'in':
            x = self.adapter_layer_norm_before(x)
        
        down = self.down_proj(x)
        
        down = self.non_linear_func(down)
        down = nn.functional.dropout(down, p=self.dropout, training=self.training)
        up = self.up_proj(down)
        
        up = up * self.scale

        if self.adapter_layernorm_option == 'out':
            up = self.adapter_layer_norm_before(up)

        if add_residual:
            output = up + residual
        else:
            output = up

        return output


# part of the code was referenced from SUPERB: https://github.com/s3prl/s3prl
# and https://github.com/wngh1187/IPET/blob/main/Speechcommands_V2/W2V2/models/W2V2.py
import pdb
import copy
import torch
import argparse
import loralib as lora
import transformers.models.whisper.modeling_whisper as whisper

from torch import nn
from .adapter import Adapter
from torch.nn import functional as F
from transformers.activations import ACT2FN
from transformers import WhisperModel, AutoFeatureExtractor

model_config = {
    "whisper_base_lora_16_conv_output": {
        "pretrained_model": "whisper_base",
        "finetune_method": "lora",
        "lora_rank": 16,
        "use_conv_outpu": True,
        "classifier_link": "https://www.dropbox.com/scl/fi/sq6w9vfrlyd0t3y5xwnnm/whisper_base_lora_16_conv_output.pt?rlkey=xtot2gi9au83xi4f3kxzwjac9&dl=0",
        "lora_link": "https://www.dropbox.com/scl/fi/46bb08ek2zbt32fmqrg21/whisper_base_lora_16_conv_output_lora.pt?rlkey=452zdg99to7ntol3rjrtgccog&dl=0"
    },
    "whisper_base_lora_8_conv_output": {
        "pretrained_model": "whisper_base",
        "finetune_method": "lora",
        "lora_rank": 8,
        "use_conv_outpu": True,
        "classifier_link": "https://www.dropbox.com/scl/fi/sq6w9vfrlyd0t3y5xwnnm/whisper_base_lora_8_conv_output.pt?rlkey=xtot2gi9au83xi4f3kxzwjac9&dl=0",
        "lora_link": "https://www.dropbox.com/scl/fi/46bb08ek2zbt32fmqrg21/whisper_base_lora_8_conv_output_lora.pt?rlkey=452zdg99to7ntol3rjrtgccog&dl=0"
    },
    "whisper_small_lora_16_conv_output": {
        "pretrained_model": "whisper_small",
        "finetune_method": "lora",
        "lora_rank": 16,
        "use_conv_outpu": True,
        "classifier_link": "https://www.dropbox.com/scl/fi/sq6w9vfrlyd0t3y5xwnnm/whisper_small_lora_16_conv_output.pt?rlkey=xtot2gi9au83xi4f3kxzwjac9&dl=0",
        "lora_link": "https://www.dropbox.com/scl/fi/46bb08ek2zbt32fmqrg21/whisper_small_lora_16_conv_output_lora.pt?rlkey=452zdg99to7ntol3rjrtgccog&dl=0"
    },
    "whisper_small_lora_8_conv_output": {
        "pretrained_model": "whisper_small",
        "finetune_method": "lora",
        "lora_rank": 8,
        "use_conv_outpu": True,
        "classifier_link": "https://www.dropbox.com/scl/fi/sq6w9vfrlyd0t3y5xwnnm/whisper_small_lora_8_conv_output.pt?rlkey=xtot2gi9au83xi4f3kxzwjac9&dl=0",
        "lora_link": "https://www.dropbox.com/scl/fi/46bb08ek2zbt32fmqrg21/whisper_small_lora_8_conv_output_lora.pt?rlkey=452zdg99to7ntol3rjrtgccog&dl=0"
    },
    "whisper_tiny_lora_16_conv_output": {
        "pretrained_model": "whisper_tiny",
        "finetune_method": "lora",
        "lora_rank": 16,
        "use_conv_outpu": True,
        "classifier_link": "https://www.dropbox.com/scl/fi/sq6w9vfrlyd0t3y5xwnnm/whisper_tiny_lora_16_conv_output.pt?rlkey=xtot2gi9au83xi4f3kxzwjac9&dl=0",
        "lora_link": "https://www.dropbox.com/scl/fi/46bb08ek2zbt32fmqrg21/whisper_tiny_lora_16_conv_output_lora.pt?rlkey=452zdg99to7ntol3rjrtgccog&dl=0"
    },
    "whisper_tiny_lora_8_conv_output": {
        "pretrained_model": "whisper_tiny",
        "finetune_method": "lora",
        "lora_rank": 8,
        "use_conv_outpu": True,
        "classifier_link": "https://www.dropbox.com/scl/fi/sq6w9vfrlyd0t3y5xwnnm/whisper_tiny_lora_8_conv_output.pt?rlkey=xtot2gi9au83xi4f3kxzwjac9&dl=0",
        "lora_link": "https://www.dropbox.com/scl/fi/46bb08ek2zbt32fmqrg21/whisper_tiny_lora_8_conv_output_lora.pt?rlkey=452zdg99to7ntol3rjrtgccog&dl=0"
    }
}

# Attach code is from huggingface
class WhisperEncoderLayer(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.embed_dim = config.d_model
        self.self_attn = whisper.WhisperAttention(
            embed_dim=self.embed_dim,
            num_heads=config.encoder_attention_heads,
            dropout=config.attention_dropout,
        )
        self.self_attn_layer_norm = nn.LayerNorm(self.embed_dim)
        self.dropout = config.dropout
        self.activation_fn = ACT2FN[config.activation_function]
        self.activation_dropout = config.activation_dropout
        self.fc1 = nn.Linear(self.embed_dim, config.encoder_ffn_dim)
        self.fc2 = nn.Linear(config.encoder_ffn_dim, self.embed_dim)
        self.final_layer_norm = nn.LayerNorm(self.embed_dim)
        self.config = config
        
        if self.config.finetune_method == "embedding_prompt":
            self.embed_prompt = nn.Parameter(torch.randn([1, self.config.embedding_prompt_dim, self.embed_dim]))
            nn.init.xavier_uniform_(self.embed_prompt)
        if self.config.finetune_method == "lora" or self.config.finetune_method == "combined":
            self.fc1 = lora.Linear(self.embed_dim, config.encoder_ffn_dim, r=config.lora_rank)
            self.fc2 = lora.Linear(config.encoder_ffn_dim, self.embed_dim, r=config.lora_rank)
            
        if self.config.finetune_method == "adapter" or self.config.finetune_method == "adapter_l" or self.config.finetune_method == "combined":
            self.adapter = Adapter(
                config, 
                d_model=self.embed_dim,
                dropout=0.1, 
                bottleneck=config.adapter_hidden_dim, 
                adapter_scalar=0.1
            )

    def forward(
        self,
        hidden_states: torch.Tensor,
        attention_mask: torch.Tensor,
        layer_head_mask: torch.Tensor,
        output_attentions: bool = False,
    ):
        """
        Args:
            hidden_states (`torch.FloatTensor`): input to the layer of shape `(seq_len, batch, embed_dim)`
            attention_mask (`torch.FloatTensor`): attention mask of size
                `(batch, 1, tgt_len, src_len)` where padding elements are indicated by very large negative values.
            layer_head_mask (`torch.FloatTensor`): mask for attention heads in a given layer of size
                `(encoder_attention_heads,)`.
            output_attentions (`bool`, *optional*):
                Whether or not to return the attentions tensors of all attention layers. See `attentions` under
                returned tensors for more detail.
        """
        if self.config.finetune_method == "embedding_prompt" or self.config.finetune_method == "combined":
            hidden_states = torch.cat((self.embed_prompt.repeat(hidden_states.size(0), 1, 1), hidden_states), dim=1)
        
        residual = hidden_states
        hidden_states = self.self_attn_layer_norm(hidden_states)
        hidden_states, attn_weights, _ = self.self_attn(
            hidden_states=hidden_states,
            attention_mask=attention_mask,
            layer_head_mask=layer_head_mask,
            output_attentions=output_attentions,
        )
        hidden_states = nn.functional.dropout(hidden_states, p=self.dropout, training=self.training)
        hidden_states = residual + hidden_states
        residual = hidden_states
        
        # Adapter
        if self.config.finetune_method == "adapter":
            adapt_h = self.adapter(hidden_states)
        
        hidden_states = self.final_layer_norm(hidden_states)
        hidden_states = self.activation_fn(self.fc1(hidden_states))
        hidden_states = nn.functional.dropout(hidden_states, p=self.activation_dropout, training=self.training)
        hidden_states = self.fc2(hidden_states)
        hidden_states = nn.functional.dropout(hidden_states, p=self.dropout, training=self.training)
        
        # Adapter
        if self.config.finetune_method == "adapter_l" or self.config.finetune_method == "combined": 
            hidden_states = hidden_states + self.adapter(hidden_states)
        
        hidden_states = residual + hidden_states
        
        if hidden_states.dtype == torch.float16 and (
            torch.isinf(hidden_states).any() or torch.isnan(hidden_states).any()
        ):
            clamp_value = torch.finfo(hidden_states.dtype).max - 1000
            hidden_states = torch.clamp(hidden_states, min=-clamp_value, max=clamp_value)
        
        # Adapter
        if self.config.finetune_method == "adapter": 
            hidden_states = hidden_states + adapt_h
        
        if self.config.finetune_method == "embedding_prompt" or self.config.finetune_method == "combined":
            hidden_states = hidden_states[:, self.config.embedding_prompt_dim:, :]
        outputs = (hidden_states,)

        if output_attentions:
            outputs += (attn_weights,)

        return outputs
   
class WhisperSER(nn.Module):
    def __init__(
        self, 
        pretrained_model:       str = "whisper_tiny",
        hidden_dim:             int = 256,
        output_class_num:       int = 4,
        finetune_method:        str = "lora",
        adapter_hidden_dim:     int = 128,
        embedding_prompt_dim:   int = 3,
        lora_rank:              int = 8,
        cache_dir:              str = ".",
        use_conv_output:        bool = True,
        enable_peft_training:   bool = True
    ):
        super(WhisperSER, self).__init__()
        # 1. We Load the model first with weights
        self.feature_extractor = AutoFeatureExtractor.from_pretrained("openai/whisper-tiny")
        self.pretrained_model = pretrained_model
        
        if self.pretrained_model == "whisper_tiny":
            self.backbone_model = WhisperModel.from_pretrained(
                "openai/whisper-tiny",
                output_hidden_states=True,
                cache_dir=cache_dir
            )
        elif self.pretrained_model == "whisper_base":
            self.backbone_model = WhisperModel.from_pretrained(
                "openai/whisper-base",
                output_hidden_states=True,
                cache_dir=cache_dir
            )
        elif self.pretrained_model == "whisper_small":
            self.backbone_model = WhisperModel.from_pretrained(
                "openai/whisper-small",
                output_hidden_states=True,
                cache_dir=cache_dir
            )
        elif self.pretrained_model == "whisper_medium":
            self.backbone_model = WhisperModel.from_pretrained(
                "openai/whisper-medium",
                output_hidden_states=True,
                cache_dir=cache_dir
            )
        self.embed_positions = copy.deepcopy(self.backbone_model.encoder.embed_positions.weight)
        self.embed_positions.requires_grad = False
        state_dict = self.backbone_model.state_dict()
        # 2. Read the model config
        self.model_config = self.backbone_model.config
        self.model_config.finetune_method        = finetune_method
        self.model_config.adapter_hidden_dim     = adapter_hidden_dim
        self.model_config.embedding_prompt_dim   = embedding_prompt_dim
        self.model_config.lora_rank              = lora_rank
        
        # 3. Config encoder layers with adapter or embedding prompt
        self.backbone_model.encoder.layers = nn.ModuleList(
            [WhisperEncoderLayer(self.model_config) for _ in range(self.model_config.encoder_layers)]
        )
        # 4. Load the weights back
        msg = self.backbone_model.load_state_dict(state_dict, strict=False)
        # 5. Freeze the weights
        for name, p in self.backbone_model.named_parameters():
            if name in msg.missing_keys and enable_peft_training: 
                p.requires_grad = True
            else: 
                p.requires_grad = False
        self.finetune_method = finetune_method
        
        # 6. Downstream models
        self.model_seq = nn.Sequential(
            nn.Conv1d(self.model_config.hidden_size, hidden_dim, 1, padding=0),
            nn.ReLU(),
            nn.Dropout(p=0.1),
            nn.Conv1d(hidden_dim, hidden_dim, 1, padding=0),
            nn.ReLU(),
            nn.Dropout(p=0.1),
            nn.Conv1d(hidden_dim, hidden_dim, 1, padding=0)
        )

        self.use_conv_output = use_conv_output
        if self.use_conv_output:
            num_layers = self.model_config.num_hidden_layers + 1  # transformer layers + input embeddings
            self.layer_weights = nn.Parameter(torch.ones(num_layers)/num_layers)
        else:
            num_layers = self.model_config.num_hidden_layers
            self.layer_weights = nn.Parameter(torch.zeros(num_layers))
        
        self.out_layer = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, output_class_num),
        )
        
        # initiate model setting
        self.init_peft_model_setting()

    def init_peft_model_setting(self):
        self.model_setting = self.finetune_method
        if self.finetune_method == "lora":
            self.model_setting += f"_{self.model_config.lora_rank}"
        elif self.finetune_method == "adapter":
            self.model_setting += f"_{self.model_config.adapter_hidden_dim}"
        elif self.finetune_method == "embedding_prompt":
            self.model_setting += f"_{self.model_config.embedding_prompt_dim}"
        if self.use_conv_output:
            self.model_setting += "_conv_output"
            
    def forward(self, x, length=None):
        # 0. check input length
        assert len(x.shape) == 2, "Input data shape wrong"
        # assert len(x[0]) <= 10 * 16000, "SER training was using 10s window frame, please crop your data to 10s"

        # 1. feature extraction and projections
        if length is not None:
            max_audio_len = length.max().detach().cpu()
            # Append to list for feature_extractor to work
            new_x = list()
            for idx in range(len(length)):
                new_x.append(x[idx].detach().cpu().numpy())
            
            # Max length is max audio len in a batch
            features = self.feature_extractor(
                new_x,
                return_tensors="pt", 
                sampling_rate=16000,
                max_length=max_audio_len
            )
            features = features.input_features.to(x.device)
        else:
            features = self.feature_extractor(
                x[0].detach().cpu(), 
                return_tensors="pt", 
                sampling_rate=16000,
                max_length=len(x[0])
            )
            features = features.input_features.to(x.device)
        
        # 2. get length and mask
        if length is not None:
            length = self.get_feat_extract_output_lengths(length.detach().cpu())
            max_len = length.max()
            # Replace positional embeddings
            self.backbone_model.encoder.embed_positions = self.backbone_model.encoder.embed_positions.from_pretrained(self.embed_positions[:max_len])
        else:
            tmp_length = self.get_feat_extract_output_lengths(len(x[0]))
            # Replace positional embeddings
            self.backbone_model.encoder.embed_positions = self.backbone_model.encoder.embed_positions.from_pretrained(self.embed_positions[:tmp_length])
            
        # 3. transformer encoding features
        features = self.backbone_model.encoder(
            features, output_hidden_states=True
        ).hidden_states
        
        # 4. stacked feature
        if self.use_conv_output:
            stacked_feature = torch.stack(features, dim=0)
        else:
            stacked_feature = torch.stack(features, dim=0)[1:]
        
        # 5. Weighted sum
        _, *origin_shape = stacked_feature.shape
        # Return transformer enc outputs [num_enc_layers, B, T, D]
        if self.use_conv_output:
            stacked_feature = stacked_feature.view(self.backbone_model.config.num_hidden_layers+1, -1)
        else:
            stacked_feature = stacked_feature.view(self.backbone_model.config.num_hidden_layers, -1)
        norm_weights = F.softmax(self.layer_weights, dim=-1)
        
        # Perform weighted average
        weighted_feature = (norm_weights.unsqueeze(-1) * stacked_feature).sum(dim=0)
        features = weighted_feature.view(*origin_shape)
        
        # 6. Pass the weighted average to point-wise 1D Conv
        # B x T x D
        features = features.transpose(1, 2)
        features = self.model_seq(features)
        features = features.transpose(1, 2)
        
        # 7. Pooling
        if length is not None:
            length = length.to(x.device)
            masks = torch.arange(features.size(1)).expand(length.size(0), -1).to(x.device) < length.unsqueeze(1)
            masks = masks.float()
            features = (features * masks.unsqueeze(-1)).sum(1) / length.unsqueeze(1)
        else:
            features = torch.mean(features, dim=1)
        
        # 8. Output predictions
        # B x D
        predicted = self.out_layer(features)
        return predicted
        
    # From huggingface
    def get_feat_extract_output_lengths(self, input_lengths):
        """
        Computes the output length of the convolutional layers
        """
        input_lengths = input_lengths // 160
        input_lengths = (input_lengths - 1) // 2 + 1
        return input_lengths

def prepare_mask(length, shape, dtype):
    # Modified from huggingface
    mask = torch.zeros(
        shape, dtype=dtype
    )
    # these two operations makes sure that all values
    # before the output lengths indices are attended to
    mask[(torch.arange(mask.shape[0]), length.cpu() - 1)] = 1
    mask = mask.flip([-1]).cumsum(-1).flip([-1]).bool()
    return mask
    
    
if __name__ == '__main__':
    
    parser = argparse.ArgumentParser(description='emo2vec finetune experiments')
    parser.add_argument(
        '--pretrained_model', 
        default='whisper_tiny',
        type=str, 
        help='finetune method: whisper_tiny, whisper_base, whisper_small'
    )
    
    parser.add_argument(
        '--finetune_method', 
        default='none',
        type=str, 
        help='finetune method: adapter, embedding prompt, input prompt'
    )
    
    parser.add_argument(
        '--adapter_hidden_dim', 
        default=128,
        type=int, 
        help='adapter dimension'
    )
    
    parser.add_argument(
        '--embedding_prompt_dim', 
        default=5,
        type=int, 
        help='prompt dimension'
    )
    
    parser.add_argument(
        '--lora_rank', 
        default=16,
        type=int, 
        help='adapter dimension'
    )
    
    args = parser.parse_args()
    model = WhisperWrapper(args).cuda()
    data = torch.zeros([1, 16000]).cuda()
    output = model(data)
    print(output.shape)


import json
import yaml
import torch
import random
import numpy as np
import pandas as pd
import torch.nn as nn
import loralib as lora
import argparse, logging
import torch.multiprocessing
import copy, time, pickle, shutil, sys, os, pdb

# from utils
from peft_ser.model.whisper import WhisperSER
from peft_ser.model.wavlm_plus import WavLMPlusSER
from peft_ser.model.wavlm_large import WavLMLargeSER
from peft_ser.model.mms import MMSSER
# from wavlm_plus import WavLMWrapper
# from wavlm_large import WavLMLargeWrapper
# from whisper import WhisperWrapper

# define logging console
import logging
logging.basicConfig(
    format='%(asctime)s %(levelname)-3s ==> %(message)s', 
    level=logging.INFO, 
    datefmt='%Y-%m-%d %H:%M:%S'
)

os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 

if __name__ == '__main__':

    # model_name = "wavlm_large"
    # model_name = "whisper_small"
    # model_name = "wavlm_plus"
    # model_name = "whisper_tiny"
    model_name = "mms-300m"
    lora_rank = 16
    
    if "whisper" in model_name:
        model = WhisperSER(
            pretrained_model=model_name,
            output_class_num=4,
            finetune_method="lora",
            lora_rank=lora_rank,
            cache_dir="/scratch1/tiantiaf/",
            use_conv_output=True,
            enable_peft_training=True
        )
    elif model_name == "wavlm_plus":
        model = WavLMPlusSER(
            pretrained_model=model_name,
            output_class_num=4,
            finetune_method="lora",
            lora_rank=lora_rank,
            cache_dir="/scratch1/tiantiaf/",
            use_conv_output=True,
            enable_peft_training=True
        )
    elif model_name == "wavlm_large":
        model = WavLMLargeSER(
            pretrained_model=model_name,
            output_class_num=4,
            finetune_method="lora",
            lora_rank=lora_rank,
            cache_dir="/scratch1/tiantiaf/",
            use_conv_output=True,
            enable_peft_training=True
        )
    elif model_name == "mms-300m":
        model = MMSSER(
            pretrained_model=model_name,
            output_class_num=4,
            finetune_method="lora",
            lora_rank=lora_rank,
            cache_dir="/scratch1/tiantiaf/",
            use_conv_output=True,
            enable_peft_training=True
        )
    # f"/project/shrikann_35/tiantiaf/effcient-ser/finetune/{model_name}/lr000025_ep30_{model.model_setting}"
    # /project/shrikann_35/tiantiaf/effcient-ser/finetune/
    model_path = f"/project/shrikann_35/tiantiaf/effcient-ser/finetune/emo4/{model_name}/lr000025_ep30_{model.model_setting}/fold_1.pt"
    lora_model_path = f"/project/shrikann_35/tiantiaf/effcient-ser/finetune/emo4/{model_name}/lr000025_ep30_{model.model_setting}/lora_fold_1.pt"
    
    save_path = f"/project/shrikann_35/tiantiaf/effcient-ser/finetune/release/{model_name}_{model.model_setting}.pt"
    lora_save_path = f"/project/shrikann_35/tiantiaf/effcient-ser/finetune/release/{model_name}_{model.model_setting}_lora.pt"
    weights = torch.load(model_path)
    weights["layer_weights"] = weights["weights"]
    del weights["weights"]
    
    filter_weights = {k: v for k, v in weights.items() if 'backbone_model' not in k and "embed_positions" not in k}
    # model.load_state_dict(torch.load(model_path))
    lora_weights = torch.load(lora_model_path)
    
    torch.save(filter_weights, save_path)
    torch.save(lora_weights, lora_save_path)
    # filter_weights.keys()
    # pdb.set_trace()


# part of the code was referenced from SUPERB: https://github.com/s3prl/s3prl
# and https://github.com/wngh1187/IPET/blob/main/Speechcommands_V2/W2V2/models/W2V2.py
import os
import pdb
import torch
import argparse
import loralib as lora
import transformers.models.wav2vec2.modeling_wav2vec2 as w2v2

from torch import nn
from .adapter import Adapter
from torch.nn import functional as F
from transformers import Wav2Vec2Model, AutoProcessor


class Wav2Vec2EncoderLayerStableLayerNorm(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.attention = w2v2.Wav2Vec2Attention(
            embed_dim=config.hidden_size,
            num_heads=config.num_attention_heads,
            dropout=config.attention_dropout,
            is_decoder=False,
        )
        self.dropout = nn.Dropout(config.hidden_dropout)
        self.layer_norm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.feed_forward = w2v2.Wav2Vec2FeedForward(config)
        self.final_layer_norm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.config = config

        if getattr(config, "adapter_attn_dim", None) is not None:
            self.adapter_layer = w2v2.Wav2Vec2AttnAdapterLayer(config)
        else:
            self.adapter_layer = None
        
        if self.config.finetune_method == "lora" or self.config.finetune_method == "combined":
            self.feed_forward.intermediate_dense    = lora.Linear(config.hidden_size, config.intermediate_size, r=config.lora_rank)
            self.feed_forward.output_dense          = lora.Linear(config.intermediate_size, config.hidden_size, r=config.lora_rank)
        
    def forward(
        self,
        hidden_states: torch.Tensor,
        attention_mask=None,
        output_attentions: bool = False,
    ):
        attn_residual = hidden_states
        hidden_states = self.layer_norm(hidden_states)
        hidden_states, attn_weights, _ = self.attention(
            hidden_states, attention_mask=attention_mask, output_attentions=output_attentions
        )
        hidden_states = self.dropout(hidden_states)
        hidden_states = attn_residual + hidden_states
        hidden_states = hidden_states + self.feed_forward(self.final_layer_norm(hidden_states))

        if self.adapter_layer is not None:
            hidden_states = hidden_states + self.adapter_layer(hidden_states)

        outputs = (hidden_states,)

        if output_attentions:
            outputs += (attn_weights,)

        return outputs

class MMSSER(nn.Module):
    def __init__(
        self, 
        pretrained_model:       str = "mms-300m",
        hidden_dim:             int = 256,
        output_class_num:       int = 4,
        finetune_method:        str = "lora",
        adapter_hidden_dim:     int = 128,
        embedding_prompt_dim:   int = 3,
        lora_rank:              int = 8,
        cache_dir:              str = ".",
        use_conv_output:        bool = True,
        enable_peft_training:   bool = True,
    ):
        super(MMSSER, self).__init__()
        # 1. We Load the model first with weights
        self.processor = AutoProcessor.from_pretrained(
            "facebook/mms-1b-all", 
            target_lang="eng", 
            cache_dir=cache_dir
        )
        
        self.backbone_model = Wav2Vec2Model.from_pretrained(
            "facebook/mms-300m",
            output_hidden_states=True,
            cache_dir=cache_dir
        )
        state_dict = self.backbone_model.state_dict()
        # 2. Read the model config
        self.model_config = self.backbone_model.config
        self.model_config.finetune_method        = finetune_method
        self.model_config.adapter_hidden_dim     = adapter_hidden_dim
        self.model_config.embedding_prompt_dim   = embedding_prompt_dim
        self.model_config.lora_rank              = lora_rank
        self.layer_norm                          = nn.LayerNorm(self.model_config.hidden_size, eps=self.model_config.layer_norm_eps)
        
        # 3. Config encoder layers with adapter or embedding prompt
        # pdb.set_trace()
        self.backbone_model.encoder.layers = nn.ModuleList([Wav2Vec2EncoderLayerStableLayerNorm(self.model_config) for i in range(self.model_config.num_hidden_layers)])
        # 4. Load the weights back
        msg = self.backbone_model.load_state_dict(state_dict, strict=False)
        # 5. Freeze the weights
        for name, p in self.backbone_model.named_parameters():
            if name in msg.missing_keys and enable_peft_training: 
                p.requires_grad = True
            else: 
                p.requires_grad = False
        self.finetune_method = finetune_method
        
        # 6. Downstream models
        self.model_seq = nn.Sequential(
            nn.Conv1d(self.model_config.hidden_size, hidden_dim, 1, padding=0),
            nn.ReLU(),
            nn.Dropout(p=0.1),
            nn.Conv1d(hidden_dim, hidden_dim, 1, padding=0),
            nn.ReLU(),
            nn.Dropout(p=0.1),
            nn.Conv1d(hidden_dim, hidden_dim, 1, padding=0)
        )
        self.use_conv_output = use_conv_output
        if use_conv_output:
            num_layers = self.model_config.num_hidden_layers + 1  # transformer layers + input embeddings
        else:
            num_layers = self.model_config.num_hidden_layers
        self.layer_weights = nn.Parameter(torch.ones(num_layers)/num_layers)
        
        self.out_layer = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, output_class_num),
        )
        
        # initiate model setting
        self.init_peft_model_setting()
        
    def init_peft_model_setting(self):
        self.model_setting = self.finetune_method
        if self.finetune_method == "lora":
            self.model_setting += f"_{self.model_config.lora_rank}"
        elif self.finetune_method == "adapter":
            self.model_setting += f"_{self.model_config.adapter_hidden_dim}"
        elif self.finetune_method == "embedding_prompt":
            self.model_setting += f"_{self.model_config.embedding_prompt_dim}"
        if self.use_conv_output:
            self.model_setting += "_conv_output"

    def forward(self, x, length=None):
        # 1. feature extraction and projections
        with torch.no_grad():
            x = self.processor(x, sampling_rate=16000, return_tensors="pt")["input_values"][0].to(x.device)
            attention_mask = torch.ones(x.shape, device=x.device, dtype=torch.long)
            x = self.backbone_model.feature_extractor(x)
            x = x.transpose(1, 2) # New version of huggingface
            x, _ = self.backbone_model.feature_projection(x) # New version of huggingface
        
        # 2. get length and mask
        if length is not None:
            for i in range(len(length)):
                attention_mask[i, length[i]:] = 0

            length = self._get_feat_extract_output_lengths(length.detach().cpu())
            length = length.to(x.device)
            
        # 3. transformer encoding features
        if length is not None:
            # compute reduced attention_mask corresponding to feature vectors
            attention_mask = self._get_feature_vector_attention_mask(
                x.shape[1], attention_mask, add_adapter=False
            )
            x = self.backbone_model.encoder(
                x, 
                attention_mask=attention_mask,
                output_hidden_states=True
            ).hidden_states
        else:
            x = self.backbone_model.encoder(
                x, output_hidden_states=True
            ).hidden_states

        # 4. stacked feature
        stacked_feature = torch.stack(x, dim=0)[:]

        # 5. Weighted sum
        # pdb.set_trace()
        _, *origin_shape = stacked_feature.shape
        # Return transformer enc outputs [num_enc_layers, B, T, D]
        if self.use_conv_output:
            stacked_feature = stacked_feature.view(self.backbone_model.config.num_hidden_layers+1, -1)
        else:
            stacked_feature = stacked_feature.view(self.backbone_model.config.num_hidden_layers, -1)
        norm_weights = F.softmax(self.layer_weights, dim=-1)
        
        # Perform weighted average
        weighted_feature = (norm_weights.unsqueeze(-1) * stacked_feature).sum(dim=0)
        weighted_feature = weighted_feature.view(*origin_shape)
        
        # 6. Pass the weighted average to point-wise 1D Conv
        # B x T x D
        features = weighted_feature.transpose(1, 2)
        features = self.model_seq(features)
        features = features.transpose(1, 2)
        
        # 7. Pooling
        if length is not None:
            length = length.to(x.device)
            masks = torch.arange(features.size(1)).expand(length.size(0), -1).to(x.device) < length.unsqueeze(1)
            masks = masks.float()
            features = (features * masks.unsqueeze(-1)).sum(1) / length.unsqueeze(1)
        else:
            features = torch.mean(features, dim=1)

        # 7. Output predictions
        # B x D
        predicted = self.out_layer(features)
        return predicted
        
    # From huggingface
    def _get_feat_extract_output_lengths(self, input_length):
        """
        Computes the output length of the convolutional layers
        """
        def _conv_out_length(input_length, kernel_size, stride):
            # 1D convolutional layer output length formula taken
            # from https://pytorch.org/docs/stable/generated/torch.nn.Conv1d.html
            return (input_length - kernel_size) // stride + 1
        for kernel_size, stride in zip(self.backbone_model.config.conv_kernel, self.backbone_model.config.conv_stride):
            input_length = _conv_out_length(input_length, kernel_size, stride)
        return input_length
    
    def _get_feature_vector_attention_mask(
        self, feature_vector_length: int, attention_mask: torch.LongTensor, add_adapter=None
    ):
        # Effectively attention_mask.sum(-1), but not inplace to be able to run
        # on inference mode.
        non_padded_lengths = attention_mask.cumsum(dim=-1)[:, -1]

        output_lengths = self._get_feat_extract_output_lengths(non_padded_lengths)
        output_lengths = output_lengths.to(torch.long)

        batch_size = attention_mask.shape[0]

        attention_mask = torch.zeros(
            (batch_size, feature_vector_length), dtype=attention_mask.dtype, device=attention_mask.device
        )
        # these two operations makes sure that all values before the output lengths idxs are attended to
        attention_mask[(torch.arange(attention_mask.shape[0], device=attention_mask.device), output_lengths - 1)] = 1
        attention_mask = attention_mask.flip([-1]).cumsum(-1).flip([-1]).bool()
        return attention_mask

def prepare_mask(length, shape, dtype):
    # Modified from huggingface
    mask = torch.zeros(
        shape, dtype=dtype
    )
    # these two operations makes sure that all values
    # before the output lengths indices are attended to
    mask[(torch.arange(mask.shape[0]), length.cpu() - 1)] = 1
    mask = mask.flip([-1]).cumsum(-1).flip([-1]).bool()
    return mask
    

def stft(x):
    return torch.stft(
        x, 
        n_fft=400, 
        hop_length=320, 
        win_length=400, 
        center=True, 
        normalized=True, 
        onesided=True,
        pad_mode='reflect',
        return_complex=True,
        window=torch.hamming_window(400).to(x.device)
    )

def istft(fea, length):
    return torch.istft(
        fea, 
        n_fft=400, 
        hop_length=320, 
        win_length=400, 
        center=True, 
        normalized=True, 
        onesided=True, 
        window=torch.hamming_window(400).to(fea.device),
        return_complex=False,
        length=length
    )
    

class BLSTM(nn.Module):
    def __init__(self,input_size,hidden_size,num_layers=1,dropout=0):
        super().__init__()
        self.blstm = nn.LSTM(
            input_size=input_size, 
            hidden_size=hidden_size, 
            batch_first=True, 
            num_layers=num_layers, 
            dropout=dropout, 
            bidirectional=True
        )

    def forward(self,x):
        # import pdb;pdb.set_trace()
        out, _ = self.blstm(x)
        out = out[:,:,:int(out.size(-1)/2)]+out[:,:,int(out.size(-1)/2):] 
        return out
    
if __name__ == '__main__':
    
    parser = argparse.ArgumentParser(description='emo2vec finetune experiments')
    parser.add_argument(
        '--finetune_method', 
        default='none',
        type=str, 
        help='finetune method: adapter, embedding prompt, input prompt'
    )
    
    parser.add_argument(
        '--adapter_hidden_dim', 
        default=128,
        type=int, 
        help='adapter dimension'
    )
    
    parser.add_argument(
        '--embedding_prompt_dim', 
        default=5,
        type=int, 
        help='adapter dimension'
    )

    parser.add_argument(
        '--lora_rank', 
        default=16,
        type=int, 
        help='lora rank'
    )

    parser.add_argument(
        '--include_layer', 
        default=0,
        type=int, 
        help='include layer'
    )

    parser.add_argument(
        '--downstream_model', 
        default="ecapa",
        type=str, 
        help='downstream models'
    )

    parser.add_argument(
        '--pretrain_model', 
        default="mms",
        type=str, 
        help='pretrained models'
    )
    
    args = parser.parse_args()
    model = MMSWrapper(args)
    # model.eval()
    data = torch.zeros([1, 16000])
    output = model(data)


import os
import pdb
import torch

import urllib
from tqdm import tqdm
from pathlib import Path

from .model.whisper import WhisperSER
from .model.wavlm_plus import WavLMPlusSER
from .model.wavlm_large import WavLMLargeSER
from .model.mms import MMSSER


model_config = {
    "whisper-base-lora-16-conv": {
        "pretrained_model": "whisper_base",
        "finetune_method": "lora",
        "lora_rank": 16,
        "use_conv_output": True,
        "classifier_link": "https://www.dropbox.com/scl/fi/sq6w9vfrlyd0t3y5xwnnm/whisper_base_lora_16_conv_output.pt?rlkey=xtot2gi9au83xi4f3kxzwjac9&dl=1",
        "lora_link": "https://www.dropbox.com/scl/fi/46bb08ek2zbt32fmqrg21/whisper_base_lora_16_conv_output_lora.pt?rlkey=452zdg99to7ntol3rjrtgccog&dl=1"
    },
    "whisper-small-lora-16-conv": {
        "pretrained_model": "whisper_small",
        "finetune_method": "lora",
        "lora_rank": 16,
        "use_conv_output": True,
        "classifier_link": "https://www.dropbox.com/scl/fi/q99rdffwnun4f62ksr6wl/whisper_small_lora_16_conv_output.pt?rlkey=1jb66t26cgh18wi4zd8yhi27s&dl=1",
        "lora_link": "https://www.dropbox.com/scl/fi/deq7cyylglk9ztf5gqr7i/whisper_small_lora_16_conv_output_lora.pt?rlkey=98ub4alelsyw5gddw6k1nwqn7&dl=1"
    },
    "whisper-tiny-lora-16-conv": {
        "pretrained_model": "whisper_tiny",
        "finetune_method": "lora",
        "lora_rank": 16,
        "use_conv_output": True,
        "classifier_link": "https://www.dropbox.com/scl/fi/ld0pymoj8iiiaen9qgpom/whisper_tiny_lora_16_conv_output.pt?rlkey=7sum57se4fmdqwhvte7eiexkc&dl=1",
        "lora_link": "https://www.dropbox.com/scl/fi/v9t8eyvsq882a0ulht80h/whisper_tiny_lora_16_conv_output_lora.pt?rlkey=1c3uzxifkm7788q61lnmkrevt&dl=1"
    },
    "wavlm-large-lora-16-conv": {
        "pretrained_model": "wavlm_large",
        "finetune_method": "lora",
        "lora_rank": 16,
        "use_conv_output": True,
        "classifier_link": "https://www.dropbox.com/scl/fi/1br3nrrnfmn0pd3kpmp5s/wavlm_large_lora_16_conv_output.pt?rlkey=jnghxe6mcz9cw78kai09k4yex&dl=1",
        "lora_link": "https://www.dropbox.com/scl/fi/4ybo1iu62t178albsbpcv/wavlm_large_lora_16_conv_output_lora.pt?rlkey=bjpjv0s66hl8z3bngzgtfzqds&dl=1"
    },
    "wavlm-plus-lora-16-conv": {
        "pretrained_model": "wavlm_plus",
        "finetune_method": "lora",
        "lora_rank": 16,
        "use_conv_output": True,
        "classifier_link": "https://www.dropbox.com/scl/fi/zlhdeoxzweksxt59d7jss/wavlm_plus_lora_16_conv_output.pt?rlkey=5trov7d6ki48bwcl25e63ffko&dl=1",
        "lora_link": "https://www.dropbox.com/scl/fi/19sb7huz9cknlybumjnfv/wavlm_plus_lora_16_conv_output_lora.pt?rlkey=8dlc3vei5cdir24wv1h1av187&dl=1"
    },
    "mms-lora-16-conv": {
        "pretrained_model": "mms-300m",
        "finetune_method": "lora",
        "lora_rank": 16,
        "use_conv_output": True,
        "classifier_link": "https://www.dropbox.com/scl/fi/mptq3vpu5b71o9azqvw8d/mms-300m_lora_16_conv_output.pt?rlkey=sj09sgy8sri0faz2eb194orwf&dl=1",
        "lora_link": "https://www.dropbox.com/scl/fi/z6uib8ivqlfty7wla8wjq/mms-300m_lora_16_conv_output_lora.pt?rlkey=nc46hvby59o1on8hiomitc916&dl=1"
    }
}

def _download_from_url(url: str, download_target: str):
    # os.makedirs(root, exist_ok=True)
    # download_target = os.path.join(root, os.path.basename(url))
    if os.path.exists(download_target) and not os.path.isfile(download_target):
        raise RuntimeError(f"{download_target} exists and is not a regular file")
        
    with urllib.request.urlopen(url) as source, open(download_target, "wb") as output:
        with tqdm(
            total=int(source.info().get("Content-Length")),
            ncols=80,
            unit="iB",
            unit_scale=True,
            unit_divisor=1024,
        ) as loop:
            while True:
                buffer = source.read(8192)
                if not buffer:
                    break
                output.write(buffer)
                loop.update(len(buffer))

def load_model(pretrained_ser_model, cache_folder: str=str(Path.cwd())):
    # /scratch1/tiantiaf/test
    # whisper_base_lora_16_conv_output
    # ser_model.load_model("whisper-base-lora-16-conv", cache_folder="/scratch1/tiantiaf/test")
    assert pretrained_ser_model in model_config.keys(), "Invalid model name"
    
    model_name = model_config[pretrained_ser_model]["pretrained_model"]
    classifier_url = model_config[pretrained_ser_model]["classifier_link"]
    lora_url = model_config[pretrained_ser_model]["lora_link"]
    finetune_method = model_config[pretrained_ser_model]["finetune_method"]
    lora_rank = model_config[pretrained_ser_model]["lora_rank"]
    use_conv_output = model_config[pretrained_ser_model]["use_conv_output"]
    
    if "whisper" in model_name:
        model = WhisperSER(
            pretrained_model=model_name,
            output_class_num=4,
            finetune_method=finetune_method,
            lora_rank=lora_rank,
            cache_dir=cache_folder,
            use_conv_output=use_conv_output,
            enable_peft_training=False
        )
    elif model_name == "wavlm_plus":
        model = WavLMPlusSER(
            pretrained_model=model_name,
            output_class_num=4,
            finetune_method=finetune_method,
            lora_rank=lora_rank,
            cache_dir=cache_folder,
            use_conv_output=use_conv_output,
            enable_peft_training=True
        )
    elif model_name == "wavlm_large":
        model = WavLMLargeSER(
            pretrained_model=model_name,
            output_class_num=4,
            finetune_method=finetune_method,
            lora_rank=lora_rank,
            cache_dir=cache_folder,
            use_conv_output=use_conv_output,
            enable_peft_training=True
        )
    elif model_name == "mms-300m":
        model = MMSSER(
            pretrained_model=model_name,
            output_class_num=4,
            finetune_method="lora",
            lora_rank=lora_rank,
            cache_dir=cache_folder,
            use_conv_output=True,
            enable_peft_training=True
        )
    # Load classifier
    save_path = str(Path.joinpath(Path(cache_folder), f"{pretrained_ser_model}.pt"))
    if not Path.exists(Path(save_path)):
        _download_from_url(url=classifier_url, download_target=save_path)
    with (open(save_path, "rb")) as fp: 
        checkpoint = torch.load(fp)
    del save_path
    model.load_state_dict(checkpoint, strict=False)

    # Load LoRa
    if finetune_method == "lora":
        lora_path = str(Path.joinpath(Path(cache_folder), f"{pretrained_ser_model}_lora.pt"))
        if not Path.exists(Path(lora_path)):
            _download_from_url(url=lora_url, download_target=lora_path)
        with (open(lora_path, "rb")) as fp: 
            checkpoint = torch.load(fp)
        model.load_state_dict(checkpoint, strict=False)
    return model


# part of the code was referenced from SUPERB: https://github.com/s3prl/s3prl
# and https://github.com/wngh1187/IPET/blob/main/Speechcommands_V2/W2V2/models/W2V2.py
import os
import pdb
import copy
import torch
import argparse
import loralib as lora
import transformers.models.wavlm.modeling_wavlm as wavlm

from torch import nn
from .adapter import Adapter
from torch.nn import functional as F
from transformers import AutoProcessor, WavLMModel

class WavLMEncoderLayer(nn.Module):
    def __init__(self, config, has_relative_position_bias: bool = True):
        super().__init__()
        self.attention = wavlm.WavLMAttention(
            embed_dim=config.hidden_size,
            num_heads=config.num_attention_heads,
            dropout=config.attention_dropout,
            num_buckets=config.num_buckets,
            max_distance=config.max_bucket_distance,
            has_relative_position_bias=has_relative_position_bias,
        )
        self.dropout = nn.Dropout(config.hidden_dropout)
        self.layer_norm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.feed_forward = wavlm.WavLMFeedForward(config)
        self.final_layer_norm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.config = config
        
        if self.config.finetune_method == "embedding_prompt" or self.config.finetune_method == "combined":
            self.embed_prompt = nn.Parameter(torch.randn([1, self.config.embedding_prompt_dim, 768]))
            nn.init.xavier_uniform_(self.embed_prompt)
        if self.config.finetune_method == "lora" or self.config.finetune_method == "combined":
            self.feed_forward.intermediate_dense    = lora.Linear(config.hidden_size, config.intermediate_size, r=config.lora_rank)
            self.feed_forward.output_dense          = lora.Linear(config.intermediate_size, config.hidden_size, r=config.lora_rank)
            
        if self.config.finetune_method == "adapter" or self.config.finetune_method == "adapter_l" or self.config.finetune_method == "combined":
            self.adapter = Adapter(
                config, 
                dropout=0.1, 
                bottleneck=config.adapter_hidden_dim, 
                adapter_scalar=0.1
            )

    def forward(self, hidden_states, attention_mask=None, position_bias=None, output_attentions=False, index=0):
        if self.config.finetune_method == "embedding_prompt" or self.config.finetune_method == "combined":
            hidden_states = torch.cat((self.embed_prompt.repeat(hidden_states.size(0), 1, 1), hidden_states), dim=1)
        
        attn_residual = hidden_states
        hidden_states, attn_weights, position_bias = self.attention(
            hidden_states,
            attention_mask=attention_mask,
            position_bias=position_bias,
            output_attentions=output_attentions,
            index=index,
        )
        hidden_states = self.dropout(hidden_states)
        hidden_states = attn_residual + hidden_states
        
        # Adapter
        if self.config.finetune_method == "adapter":
            adapt_h = self.adapter(hidden_states)

        hidden_states = self.layer_norm(hidden_states)
        hidden_states = hidden_states + self.feed_forward(hidden_states)
        if self.config.finetune_method == "adapter": 
            hidden_states = hidden_states + adapt_h
        if self.config.finetune_method == "adapter_l" or self.config.finetune_method == "combined": 
            hidden_states = hidden_states + self.adapter(hidden_states)
        hidden_states = self.final_layer_norm(hidden_states)
        if self.config.finetune_method == "embedding_prompt" or self.config.finetune_method == "combined":
            hidden_states = hidden_states[:, self.config.embedding_prompt_dim:, :]
        outputs = (hidden_states, position_bias)

        if output_attentions:
            outputs += (attn_weights,)

        return outputs


class WavLMEncoderLayerStableLayerNorm(nn.Module):
    def __init__(self, config, has_relative_position_bias: bool = True):
        super().__init__()
        self.attention = wavlm.WavLMAttention(
            embed_dim=config.hidden_size,
            num_heads=config.num_attention_heads,
            dropout=config.attention_dropout,
            num_buckets=config.num_buckets,
            max_distance=config.max_bucket_distance,
            has_relative_position_bias=has_relative_position_bias,
        )
        self.dropout = nn.Dropout(config.hidden_dropout)
        self.layer_norm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.feed_forward = wavlm.WavLMFeedForward(config)
        self.final_layer_norm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.config = config

        if self.config.finetune_method == "lora":
            self.feed_forward.intermediate_dense    = lora.Linear(config.hidden_size, config.intermediate_size, r=config.lora_rank)
            self.feed_forward.output_dense          = lora.Linear(config.intermediate_size, config.hidden_size, r=config.lora_rank)
        
    def forward(self, hidden_states, attention_mask=None, position_bias=None, output_attentions=False):
        attn_residual = hidden_states
        hidden_states = self.layer_norm(hidden_states)
        hidden_states, attn_weights, position_bias = self.attention(
            hidden_states,
            attention_mask=attention_mask,
            position_bias=position_bias,
            output_attentions=output_attentions,
        )
        hidden_states = self.dropout(hidden_states)
        hidden_states = attn_residual + hidden_states
        hidden_states = hidden_states + self.feed_forward(self.final_layer_norm(hidden_states))

        outputs = (hidden_states, position_bias)

        if output_attentions:
            outputs += (attn_weights,)

        return outputs


   
class WavLMLargeSER(nn.Module):
    def __init__(
        self, 
        pretrained_model:       str = "wavlm_plus",
        hidden_dim:             int = 256,
        output_class_num:       int = 4,
        finetune_method:        str = "lora",
        adapter_hidden_dim:     int = 128,
        embedding_prompt_dim:   int = 3,
        lora_rank:              int = 8,
        cache_dir:              str = ".",
        use_conv_output:        bool = True,
        enable_peft_training:   bool = True
    ):
        super(WavLMLargeSER, self).__init__()
        # 1. We Load the model first with weights
        self.processor = AutoProcessor.from_pretrained("facebook/mms-1b-all")
        self.backbone_model = WavLMModel.from_pretrained(
            "microsoft/wavlm-large",
            output_hidden_states=True,
            cache_dir=cache_dir
        )
        state_dict      = copy.deepcopy(self.backbone_model.state_dict())
        # 2. Read the model config
        self.model_config = self.backbone_model.config
        self.model_config.finetune_method        = finetune_method
        self.model_config.adapter_hidden_dim     = adapter_hidden_dim
        self.model_config.embedding_prompt_dim   = embedding_prompt_dim
        self.model_config.lora_rank              = lora_rank
        
        # 3. Config encoder layers with adapter or embedding prompt
        self.backbone_model.encoder.layers = nn.ModuleList(
            [WavLMEncoderLayerStableLayerNorm(self.model_config, has_relative_position_bias=(i == 0)) for i in range(self.model_config.num_hidden_layers)]
        )

        # 4. Load the weights back
        msg = self.backbone_model.load_state_dict(state_dict, strict=False)
        # 5. Freeze the weights
        for name, p in self.backbone_model.named_parameters():
            if name in msg.missing_keys and enable_peft_training: 
                p.requires_grad = True
            else: 
                p.requires_grad = False
        self.finetune_method = finetune_method
        
        # 6. Downstream models
        self.model_seq = nn.Sequential(
            nn.Conv1d(self.model_config.hidden_size, hidden_dim, 1, padding=0),
            nn.ReLU(),
            nn.Dropout(p=0.1),
            nn.Conv1d(hidden_dim, hidden_dim, 1, padding=0),
            nn.ReLU(),
            nn.Dropout(p=0.1),
            nn.Conv1d(hidden_dim, hidden_dim, 1, padding=0)
        )

        self.use_conv_output = use_conv_output
        if use_conv_output:
            num_layers = self.model_config.num_hidden_layers + 1  # transformer layers + input embeddings
        else:
            num_layers = self.model_config.num_hidden_layers

        self.layer_weights = nn.Parameter(torch.ones(num_layers)/num_layers)
        
        self.out_layer = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, output_class_num),
        )
        # initiate model setting
        self.init_peft_model_setting()

    def init_peft_model_setting(self):
        self.model_setting = self.finetune_method
        if self.finetune_method == "lora":
            self.model_setting += f"_{self.model_config.lora_rank}"
        elif self.finetune_method == "adapter":
            self.model_setting += f"_{self.model_config.adapter_hidden_dim}"
        elif self.finetune_method == "embedding_prompt":
            self.model_setting += f"_{self.model_config.embedding_prompt_dim}"
        if self.use_conv_output:
            self.model_setting += "_conv_output"

    def forward(self, x, length=None):
        # 1. feature extraction and projections
        with torch.no_grad():
            x = self.processor(x, sampling_rate=16000, return_tensors="pt")["input_values"][0].to(x.device)
            attention_mask = torch.ones(x.shape, device=x.device, dtype=torch.long)
            x = self.backbone_model.feature_extractor(x)
            x = x.transpose(1, 2) # New version of huggingface
            x, _ = self.backbone_model.feature_projection(x) # New version of huggingface
            
        # 2. get length and mask
        if length is not None:
            for i in range(len(length)):
                attention_mask[i, length[i]:] = 0

            length = self._get_feat_extract_output_lengths(length.detach().cpu())
            length = length.to(x.device)
            
        # 3. transformer encoding features
        if length is not None:
            # compute reduced attention_mask corresponding to feature vectors
            attention_mask = self._get_feature_vector_attention_mask(
                x.shape[1], attention_mask, add_adapter=False
            )
            x = self.backbone_model.encoder(
                x, 
                attention_mask=attention_mask,
                output_hidden_states=True
            ).hidden_states
        else:
            x = self.backbone_model.encoder(
                x, output_hidden_states=True
            ).hidden_states
        
        # 4. stacked feature
        stacked_feature = torch.stack(x, dim=0)[:]
        
        # 5. Weighted sum
        _, *origin_shape = stacked_feature.shape
        # Return transformer enc outputs [num_enc_layers, B, T, D]
        if self.use_conv_output:
            stacked_feature = stacked_feature.view(self.backbone_model.config.num_hidden_layers+1, -1)
        else:
            stacked_feature = stacked_feature.view(self.backbone_model.config.num_hidden_layers, -1)
        norm_weights = F.softmax(self.layer_weights, dim=-1)
        
        # Perform weighted average
        weighted_feature = (norm_weights.unsqueeze(-1) * stacked_feature).sum(dim=0)
        weighted_feature = weighted_feature.view(*origin_shape)
        
        # 6. Pass the weighted average to point-wise 1D Conv
        # B x T x D
        features = weighted_feature.transpose(1, 2)
        features = self.model_seq(features)
        features = features.transpose(1, 2)
        
        # 7. Pooling
        if length is not None:
            length = length.to(x.device)
            masks = torch.arange(features.size(1)).expand(length.size(0), -1).to(x.device) < length.unsqueeze(1)
            masks = masks.float()
            features = (features * masks.unsqueeze(-1)).sum(1) / length.unsqueeze(1)
        else:
            features = torch.mean(features, dim=1)
        
        # 7. Output predictions
        # B x D
        predicted = self.out_layer(features)
        return predicted
        
    # From huggingface
    def _get_feat_extract_output_lengths(self, input_length):
        """
        Computes the output length of the convolutional layers
        """
        def _conv_out_length(input_length, kernel_size, stride):
            # 1D convolutional layer output length formula taken
            # from https://pytorch.org/docs/stable/generated/torch.nn.Conv1d.html
            return (input_length - kernel_size) // stride + 1
        for kernel_size, stride in zip(self.backbone_model.config.conv_kernel, self.backbone_model.config.conv_stride):
            input_length = _conv_out_length(input_length, kernel_size, stride)
        return input_length

    def _get_feature_vector_attention_mask(
        self, feature_vector_length: int, attention_mask: torch.LongTensor, add_adapter=None
    ):
        # Effectively attention_mask.sum(-1), but not inplace to be able to run
        # on inference mode.
        non_padded_lengths = attention_mask.cumsum(dim=-1)[:, -1]

        output_lengths = self._get_feat_extract_output_lengths(non_padded_lengths)
        output_lengths = output_lengths.to(torch.long)

        batch_size = attention_mask.shape[0]

        attention_mask = torch.zeros(
            (batch_size, feature_vector_length), dtype=attention_mask.dtype, device=attention_mask.device
        )
        # these two operations makes sure that all values before the output lengths idxs are attended to
        attention_mask[(torch.arange(attention_mask.shape[0], device=attention_mask.device), output_lengths - 1)] = 1
        attention_mask = attention_mask.flip([-1]).cumsum(-1).flip([-1]).bool()
        return attention_mask

def prepare_mask(length, shape, dtype):
    # Modified from huggingface
    mask = torch.zeros(
        shape, dtype=dtype
    )
    # these two operations makes sure that all values
    # before the output lengths indices are attended to
    mask[(torch.arange(mask.shape[0]), length.cpu() - 1)] = 1
    mask = mask.flip([-1]).cumsum(-1).flip([-1]).bool()
    return mask


def stft(x):
    return torch.stft(
        x, 
        n_fft=400, 
        hop_length=320, 
        win_length=400, 
        center=True, 
        normalized=True, 
        onesided=True,
        pad_mode='reflect',
        return_complex=True,
        window=torch.hamming_window(400).to(x.device)
    )
    
    
if __name__ == '__main__':
    
    parser = argparse.ArgumentParser(description='emo2vec finetune experiments')
    parser.add_argument(
        '--finetune_method', 
        default='none',
        type=str, 
        help='finetune method: adapter, embedding prompt, input prompt'
    )
    
    parser.add_argument(
        '--adapter_hidden_dim', 
        default=128,
        type=int, 
        help='adapter dimension'
    )
    
    parser.add_argument(
        '--embedding_prompt_dim', 
        default=5,
        type=int, 
        help='adapter dimension'
    )
    
    args = parser.parse_args()
    model = WavLMWrapper(args)
    data = torch.zeros([1, 16000])
    output = model(data)
    print(output.shape)


import os
import platform
import sys

import pkg_resources
from setuptools import find_packages, setup

requirements = []
if sys.platform.startswith("linux") and platform.machine() == "x86_64":
    requirements.append("triton==2.0.0")
setup(
    name="peft-ser",
    py_modules=["peft_ser"],
    version="0.0.5",
    description="Parameter Efficient Fine-tuning on Speech Emotion Recognition.",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    readme="README.md",
    python_requires=">=3.8",
    author="Tiantian Feng",
    license="BSD",
    packages=find_packages(exclude=["tests*"]),
    install_requires=requirements
    + [
        str(r)
        for r in pkg_resources.parse_requirements(
            open(os.path.join(os.path.dirname(__file__), "requirements.txt"))
        )
    ],
    entry_points={
        "console_scripts": ["whisper=whisper.transcribe:cli"],
    },
    include_package_data=True,
    extras_require={"dev": ["pytest", "scipy", "black", "flake8", "isort"]},
)



# part of the code was referenced from SUPERB: https://github.com/s3prl/s3prl
# and https://github.com/wngh1187/IPET/blob/main/Speechcommands_V2/W2V2/models/W2V2.py
import os
import pdb
import copy
import torch
import argparse
import numpy as np
import loralib as lora
import transformers.models.wav2vec2.modeling_wav2vec2 as w2v2

from functools import lru_cache
from torchaudio.compliance import kaldi

from torch import nn
from adapter import Adapter
from collections import OrderedDict
from typing import Optional, Callable
from torch.nn import functional as F
from torch.nn.functional import normalize
from transformers import Wav2Vec2Model, Wav2Vec2Config, Wav2Vec2Processor, AutoProcessor, WavLMModel, WhisperModel, AutoFeatureExtractor


class Wav2Vec2EncoderLayer(nn.Module):
    def __init__(
        self, 
        config, 
        i
    ):
        super().__init__()
        self.attention = w2v2.Wav2Vec2Attention(
            embed_dim=config.hidden_size,
            num_heads=config.num_attention_heads,
            dropout=config.attention_dropout,
            is_decoder=False,
        )
        self.dropout = nn.Dropout(config.hidden_dropout)
        self.layer_norm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.feed_forward = w2v2.Wav2Vec2FeedForward(config)
        self.final_layer_norm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.config = config
        
        if self.config.finetune_method == "embedding_prompt" or self.config.finetune_method == "combined":
            self.embed_prompt = nn.Parameter(torch.randn([1, self.config.embedding_prompt_dim, 768]))
            nn.init.xavier_uniform_(self.embed_prompt)
        if self.config.finetune_method == "lora" or self.config.finetune_method == "combined":
            self.feed_forward.intermediate_dense    = lora.Linear(config.hidden_size, config.intermediate_size, r=config.lora_rank)
            self.feed_forward.output_dense          = lora.Linear(config.intermediate_size, config.hidden_size, r=config.lora_rank)
            
        if self.config.finetune_method == "adapter" or self.config.finetune_method == "adapter_l" or self.config.finetune_method == "combined":
            self.adapter = Adapter(
                config, 
                dropout=0.1, 
                bottleneck=config.adapter_hidden_dim, 
                adapter_scalar=0.1
            )
        self.i = i

    def forward(self, hidden_states, attention_mask=None, output_attentions=False):
        if self.config.finetune_method == "embedding_prompt" or self.config.finetune_method == "combined":
            hidden_states = torch.cat((self.embed_prompt.repeat(hidden_states.size(0), 1, 1), hidden_states), dim=1)
        attn_residual = hidden_states
        
        hidden_states, attn_weights, _ = self.attention(
            hidden_states, attention_mask=attention_mask, output_attentions=output_attentions
        )
        hidden_states = self.dropout(hidden_states)
        hidden_states = attn_residual + hidden_states

        # Adapter
        if self.config.finetune_method == "adapter":
            adapt_h = self.adapter(hidden_states)

        hidden_states = self.layer_norm(hidden_states)
        hidden_states = hidden_states + self.feed_forward(hidden_states) 
        # Adapter
        if self.config.finetune_method == "adapter": 
            hidden_states = hidden_states+ adapt_h
        if self.config.finetune_method == "adapter_l" or self.config.finetune_method == "combined": 
            hidden_states = hidden_states + self.adapter(hidden_states)
            
        hidden_states = self.final_layer_norm(hidden_states)
        if self.config.finetune_method == "embedding_prompt" or self.config.finetune_method == "combined":
            hidden_states = hidden_states[:, self.config.embedding_prompt_dim:, :]

        outputs = (hidden_states,)

        if output_attentions:
            outputs += (attn_weights,)
        return outputs

class Wav2VecWrapper(nn.Module):
    def __init__(
        self, 
        args, 
        hidden_dim=256,
        output_class_num=4
    ):
        super(Wav2VecWrapper, self).__init__()
        # 1. We Load the model first with weights
        self.args = args
        self.backbone_model = Wav2Vec2Model.from_pretrained(
            "facebook/wav2vec2-base-960h",
            output_hidden_states=True
        )
        state_dict = self.backbone_model.state_dict()
        # 2. Read the model config
        self.model_config = self.backbone_model.config
        self.model_config.finetune_method        = args.finetune_method
        self.model_config.adapter_hidden_dim     = args.adapter_hidden_dim
        self.model_config.embedding_prompt_dim   = args.embedding_prompt_dim
        self.model_config.lora_rank              = args.lora_rank
        
        # 3. Config encoder layers with adapter or embedding prompt
        # pdb.set_trace()
        self.backbone_model.encoder.layers = nn.ModuleList([Wav2Vec2EncoderLayer(self.model_config, i) for i in range(self.model_config.num_hidden_layers)])
        # 4. Load the weights back
        msg = self.backbone_model.load_state_dict(state_dict, strict=False)
        # 5. Freeze the weights
        if self.args.finetune_method == "adapter" or self.args.finetune_method == "adapter_l" or self.args.finetune_method == "embedding_prompt" or self.args.finetune_method == "finetune" or self.args.finetune_method == "lora" or self.args.finetune_method == "combined":
            for name, p in self.backbone_model.named_parameters():
                if name in msg.missing_keys: p.requires_grad = True
                else: p.requires_grad = False
        self.finetune_method = self.args.finetune_method
        
        # 6. Downstream models
        self.model_seq = nn.Sequential(
            nn.Conv1d(self.model_config.hidden_size, hidden_dim, 1, padding=0),
            nn.ReLU(),
            nn.Dropout(p=0.1),
            nn.Conv1d(hidden_dim, hidden_dim, 1, padding=0),
            nn.ReLU(),
            nn.Dropout(p=0.1),
            nn.Conv1d(hidden_dim, hidden_dim, 1, padding=0)
        )
        self.weights = nn.Parameter(torch.zeros(self.model_config.num_hidden_layers))
        
        self.out_layer = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, output_class_num),
        )
        
    def forward(self, x, length=None):
        # 1. feature extraction and projections
        with torch.no_grad():
            x = self.backbone_model.feature_extractor(x)
            x = x.transpose(1, 2) # New version of huggingface
            x, _ = self.backbone_model.feature_projection(x) # New version of huggingface
        
        # 2. get length and mask
        if length is not None:
            length = self.get_feat_extract_output_lengths(length.detach().cpu())
            length = length.to(x.device)
            
        # 3. transformer encoding features
        x = self.backbone_model.encoder(
            x, output_hidden_states=True
        ).hidden_states
        
        # 4. stacked feature
        stacked_feature = torch.stack(x, dim=0)[1:]
        
        # 5. Weighted sum
        _, *origin_shape = stacked_feature.shape
        # Return transformer enc outputs [num_enc_layers, B, T, D]
        stacked_feature = stacked_feature.view(self.backbone_model.config.num_hidden_layers, -1)
        norm_weights = F.softmax(self.weights, dim=-1)
        
        # Perform weighted average
        weighted_feature = (norm_weights.unsqueeze(-1) * stacked_feature).sum(dim=0)
        features = weighted_feature.view(*origin_shape)
        
        # 6. Pass the weighted average to point-wise 1D Conv
        # B x T x D
        features = features.transpose(1, 2)
        features = self.model_seq(features)
        features = features.transpose(1, 2)
        
        # 7. Pooling
        if length is not None:
            masks = torch.arange(features.size(1)).expand(length.size(0), -1).to(x.device) < length.unsqueeze(1)
            masks = masks.float()
            features = (features * masks.unsqueeze(-1)).sum(1) / length.unsqueeze(1)
        else:
            features = torch.mean(features, dim=1)
        
        # 8. Output predictions
        # B x D
        predicted = self.out_layer(features)
        return predicted
        
    # From huggingface
    def get_feat_extract_output_lengths(self, input_length):
        """
        Computes the output length of the convolutional layers
        """
        def _conv_out_length(input_length, kernel_size, stride):
            # 1D convolutional layer output length formula taken
            # from https://pytorch.org/docs/stable/generated/torch.nn.Conv1d.html
            return (input_length - kernel_size) // stride + 1
        for kernel_size, stride in zip(self.backbone_model.config.conv_kernel, self.backbone_model.config.conv_stride):
            input_length = _conv_out_length(input_length, kernel_size, stride)
        return input_length

def prepare_mask(length, shape, dtype):
    # Modified from huggingface
    mask = torch.zeros(
        shape, dtype=dtype
    )
    # these two operations makes sure that all values
    # before the output lengths indices are attended to
    mask[(torch.arange(mask.shape[0]), length.cpu() - 1)] = 1
    mask = mask.flip([-1]).cumsum(-1).flip([-1]).bool()
    return mask
    
    
if __name__ == '__main__':
    
    parser = argparse.ArgumentParser(description='emo2vec finetune experiments')
    parser.add_argument(
        '--finetune_method', 
        default='none',
        type=str, 
        help='finetune method: adapter, embedding prompt, input prompt'
    )
    
    parser.add_argument(
        '--adapter_hidden_dim', 
        default=128,
        type=int, 
        help='adapter dimension'
    )
    
    parser.add_argument(
        '--embedding_prompt_dim', 
        default=5,
        type=int, 
        help='adapter dimension'
    )
    
    args = parser.parse_args()
    model = Wav2VecWrapper(args)
    data = torch.zeros([1, 16000])
    output = model(data)
    print(output.shape)